/*
 * This file is a part of the CaosDB Project.
 *
 * Copyright (C) 2021 Timm Fitschen <t.fitschen@indiscale.com>
 * Copyright (C) 2021 IndiScale GmbH <info@indiscale.com>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 */
#include "caosdb/configuration.h"
#include "boost/iterator/iterator_facade.hpp"          // for iterator_facad...
#include "boost/json/impl/object.hpp"                  // for object::at
#include "boost/json/string.hpp"                       // for string
#include "boost/json/string_view.hpp"                  // for string_view
#include "boost/log/core/record.hpp"                   // for record
#include "boost/log/detail/attachable_sstream_buf.hpp" // for basic_ostring...
#include "boost/log/sources/record_ostream.hpp"        // for basic_record_o...
#include "boost/preprocessor/seq/limits/enum_256.hpp"  // for BOOST_PP_SEQ_E...
#include "boost/preprocessor/seq/limits/size_256.hpp"  // for BOOST_PP_SEQ_S...
#include "caosdb/authentication.h"                     // for Authenticator
#include "caosdb/connection.h"                         // for ConnectionManager
#include "caosdb/constants.h"                          // for LIBCAOSDB_CONF...
#include "caosdb/exceptions.h"                         // for ConfigurationE...
#include "caosdb/log_level.h"                          // for CAOSDB_DEFAULT...
#include "caosdb/status_code.h"                        // for StatusCode
#include "caosdb/utility.h"                            // for get_home_direc...
#include <bits/exception.h>                            // for exception
#include <cassert>                                     // for assert
#include <cstdlib>                                     // for getenv
#include <grpcpp/security/credentials.h>               // for SslCredentials
#include <iterator>                                    // for next
#include <map>                                         // for map
#include <stdexcept>                                   // for out_of_range
#include <string>                                      // for string, operator+

namespace caosdb::configuration {
using boost::filesystem::exists;
using boost::filesystem::path;
using boost::json::object;
using boost::json::value;
using caosdb::authentication::Authenticator;
using caosdb::authentication::PlainPasswordAuthenticator;
using caosdb::connection::ConnectionManager;
using caosdb::exceptions::ConfigurationError;
using caosdb::logging::ConsoleSinkConfiguration;
using caosdb::logging::FileSinkConfiguration;
using caosdb::logging::LoggingConfiguration;
using caosdb::logging::SinkConfiguration;
using caosdb::logging::SyslogSinkConfiguration;
using caosdb::utility::get_home_directory;
using caosdb::utility::load_json_file;
using caosdb::utility::load_string_file;
using grpc::InsecureChannelCredentials;
using grpc::SslCredentials;
using grpc::SslCredentialsOptions;

PemFileCertificateProvider::PemFileCertificateProvider(const path &path) {
  this->certificate_provider = load_string_file(path);
}

auto PemFileCertificateProvider::GetCertificatePem() const -> std::string {
  return this->certificate_provider;
}

PemCertificateProvider::PemCertificateProvider(
  const std::string &certificate_provider) {
  this->certificate_provider = certificate_provider;
}

auto PemCertificateProvider::GetCertificatePem() const -> std::string {
  return this->certificate_provider;
}

ConnectionConfiguration::ConnectionConfiguration(const std::string &host,
                                                 int port) {
  this->host = host;
  this->port = port;
}

auto ConnectionConfiguration::GetHost() const -> std::string {
  return this->host;
}

auto ConnectionConfiguration::GetPort() const -> int { return this->port; }

auto operator<<(std::ostream &out, const ConnectionConfiguration &configuration)
  -> std::ostream & {
  out << configuration.ToString();
  return out;
}

InsecureConnectionConfiguration::InsecureConnectionConfiguration(
  const std::string &host, int port)
  : ConnectionConfiguration(host, port) {
  this->credentials = InsecureChannelCredentials();
}

auto InsecureConnectionConfiguration::GetChannelCredentials() const
  -> std::shared_ptr<ChannelCredentials> {
  return this->credentials;
}

auto InsecureConnectionConfiguration::ToString() const -> std::string {
  return "InsecureConnectionConfiguration(" + this->GetHost() + "," +
         std::to_string(this->GetPort()) + ")";
}

TlsConnectionConfiguration::TlsConnectionConfiguration(const std::string &host,
                                                       int port)
  : ConnectionConfiguration(host, port) {
  SslCredentialsOptions options;
  this->credentials = SslCredentials(options);
}

TlsConnectionConfiguration::TlsConnectionConfiguration(
  const std::string &host, int port,
  const CertificateProvider &certificate_provider)
  : ConnectionConfiguration(host, port) {
  SslCredentialsOptions options;
  options.pem_root_certs = certificate_provider.GetCertificatePem();
  this->credentials = SslCredentials(options);
}

TlsConnectionConfiguration::TlsConnectionConfiguration(
  const std::string &host, int port, const Authenticator &authenticator)
  : ConnectionConfiguration(host, port) {

  SslCredentialsOptions options;
  this->credentials = grpc::CompositeChannelCredentials(
    SslCredentials(options), authenticator.GetCallCredentials());
}

TlsConnectionConfiguration::TlsConnectionConfiguration(
  const std::string &host, int port,
  const CertificateProvider &certificate_provider,
  const Authenticator &authenticator)
  : ConnectionConfiguration(host, port) {

  SslCredentialsOptions options;
  options.pem_root_certs = certificate_provider.GetCertificatePem();
  this->credentials = grpc::CompositeChannelCredentials(
    SslCredentials(options), authenticator.GetCallCredentials());
}

auto TlsConnectionConfiguration::GetChannelCredentials() const
  -> std::shared_ptr<ChannelCredentials> {
  return this->credentials;
}

auto TlsConnectionConfiguration::ToString() const -> std::string {
  return "TlsConnectionConfiguration(" + this->GetHost() + "," +
         std::to_string(this->GetPort()) + "," + this->certificate_provider +
         ")";
}

auto ConnectionConfigurationHelper::CreateCertificateProvider(
  const object &from) const -> std::unique_ptr<CertificateProvider> {
  std::unique_ptr<CertificateProvider> certificate_provider;
  if (from.contains("server_certificate_path")) {
    const value &path_str = from.at("server_certificate_path");
    assert(path_str.is_string() == true);
    const path certificate_file = path(path_str.as_string().c_str());
    if (!exists(certificate_file)) {
      throw ConfigurationError(
        "File does not exist (server_certificate_path): " +
        certificate_file.string());
    }
    certificate_provider = std::make_unique<PemFileCertificateProvider>(
      path(path_str.as_string().c_str()));
  }
  return certificate_provider;
}

auto ConnectionConfigurationHelper::CreateAuthenticator(
  const object &from) const -> std::unique_ptr<Authenticator> {
  std::unique_ptr<Authenticator> authenticator;
  if (from.contains("authentication")) {
    assert(from.at("authentication").is_object());
    auto authentication = from.at("authentication").as_object();
    auto type = std::string("plain");
    if (authentication.contains("type")) {
      assert(authentication.at("type").is_string());
      type = std::string(authentication.at("type").as_string().c_str());
    }
    if (type == "plain") {
      assert(authentication.contains("username"));
      auto username = authentication.at("username");
      assert(username.is_string());

      assert(authentication.contains("password"));
      auto password = authentication.at("password");
      assert(password.is_string());

      authenticator = std::make_unique<PlainPasswordAuthenticator>(
        std::string(username.as_string().c_str()),
        std::string(password.as_string().c_str()));
    } else {
      throw ConfigurationError("Unknow authentication type: '" + type + "'.");
    }
  }
  return authenticator;
}

auto ConnectionConfigurationHelper::CreateConnectionConfiguration(
  const bool tls, const std::string &host, const int port,
  const CertificateProvider *certificate_provider,
  const Authenticator *authenticator) const
  -> std::unique_ptr<ConnectionConfiguration> {
  if (tls) {
    if (certificate_provider != nullptr && authenticator != nullptr) {
      // authenticated and special certificate
      return std::make_unique<TlsConnectionConfiguration>(
        host, port, *certificate_provider, *authenticator);
    } else if (certificate_provider != nullptr) {
      // unauthenticated, special certificate
      return std::make_unique<TlsConnectionConfiguration>(
        host, port, *certificate_provider);
    } else if (authenticator != nullptr) {
      // authenticated, no special certificate
      return std::make_unique<TlsConnectionConfiguration>(host, port,
                                                          *authenticator);
    }
    // unauthenticated, no special certificate
    return std::make_unique<TlsConnectionConfiguration>(host, port);
  } else {
    return std::make_unique<InsecureConnectionConfiguration>(host, port);
  }
}

auto ConnectionConfigurationHelper::IsTls(const object &from) const -> bool {
  bool tls = true;
  if (from.contains("tls")) {
    auto tls_switch = from.at("tls");
    assert(tls_switch.is_bool());
    tls = tls_switch.as_bool();
  }
  return tls;
}

auto ConnectionConfigurationHelper::CreateConnectionConfiguration(
  const object &from) const -> std::unique_ptr<ConnectionConfiguration> {
  assert(from.contains("host"));
  const auto &host = from.at("host");
  assert(host.is_string());

  assert(from.contains("port"));
  const auto &port = from.at("port");
  assert(port.is_int64());

  auto tls = IsTls(from);

  auto certificate_provider = CreateCertificateProvider(from);

  auto authenticator = CreateAuthenticator(from);

  return CreateConnectionConfiguration(
    tls, std::string(host.as_string().c_str()),
    static_cast<int>(port.as_int64()), certificate_provider.get(),
    authenticator.get());
}

auto LoggingConfigurationHelper::CreateConsoleSinkConfiguration(
  const object & /*from*/, const std::string &name, int level) const
  -> std::shared_ptr<caosdb::logging::SinkConfiguration> {
  auto result = std::make_shared<ConsoleSinkConfiguration>(name, level);
  return result;
}

auto LoggingConfigurationHelper::CreateSyslogSinkConfiguration(
  const object & /*from*/, const std::string &name, int level) const
  -> std::shared_ptr<caosdb::logging::SinkConfiguration> {
  auto result = std::make_shared<SyslogSinkConfiguration>(name, level);
  return result;
}

auto LoggingConfigurationHelper::CreateFileSinkConfiguration(
  const object &from, const std::string &name, int level) const
  -> std::shared_ptr<caosdb::logging::SinkConfiguration> {
  auto result = std::make_shared<FileSinkConfiguration>(name, level);
  if (from.contains("directory")) {
    result->SetDirectory(from.at("directory").as_string().c_str());
  }
  return result;
}

auto LoggingConfigurationHelper::CreateSinkConfiguration(
  const object &from, const std::string &name, int default_level) const
  -> std::shared_ptr<caosdb::logging::SinkConfiguration> {
  assert(from.contains("destination"));
  const auto &destination =
    std::string(from.at("destination").as_string().c_str());
  int level = from.contains("level")
                ? static_cast<int>(from.at("level").as_int64())
                : default_level;

  if (destination == "file") {
    return CreateFileSinkConfiguration(from, name, level);
  } else if (destination == "console") {
    return CreateConsoleSinkConfiguration(from, name, level);
  } else if (destination == "syslog") {
    return CreateSyslogSinkConfiguration(from, name, level);
  } else {
    throw ConfigurationError("Unknown sink destination: " + destination);
  }
}

auto LoggingConfigurationHelper::CreateLoggingConfiguration(
  const object &from) const -> LoggingConfiguration {
  auto default_level_str = from.contains("level")
                             ? std::string(from.at("level").as_string().c_str())
                             : "";
  int default_level = 0;
  static std::map<std::string, int> log_level_names = {
    {"", CAOSDB_DEFAULT_LOG_LEVEL},    {"off", CAOSDB_LOG_LEVEL_OFF},
    {"fatal", CAOSDB_LOG_LEVEL_FATAL}, {"error", CAOSDB_LOG_LEVEL_ERROR},
    {"warn", CAOSDB_LOG_LEVEL_WARN},   {"info", CAOSDB_LOG_LEVEL_INFO},
    {"debug", CAOSDB_LOG_LEVEL_DEBUG}, {"trace", CAOSDB_LOG_LEVEL_TRACE},
    {"all", CAOSDB_LOG_LEVEL_ALL}};

  try {
    default_level = CAOSDB_DEFAULT_LOG_LEVEL;
  } catch (const std::out_of_range &exc) {
    throw ConfigurationError("Unknown log level: " + default_level_str);
  }

  auto result = LoggingConfiguration(default_level);
  if (default_level == CAOSDB_LOG_LEVEL_OFF) {
    return result;
  }

  const auto &sinks = from.at("sinks").as_object();
  if (!sinks.empty()) {
    const auto *elem = sinks.begin();

    while (elem != sinks.end()) {
      result.AddSink(CreateSinkConfiguration(
        elem->value().as_object(), elem->key().to_string(), default_level));
      elem = std::next(elem);
    }
  }

  return result;
}

auto ConfigurationManager::mReset() noexcept -> int {
  try {
    mClear();
    InitializeDefaults();
    return StatusCode::SUCCESS;
  } catch (const caosdb::exceptions::GenericException &exc) {
    return exc.GetCode();
  } catch (const std::exception &exc) {
    CAOSDB_LOG_ERROR(logger_name)
      << "Unknown error during the reset of the ConfigurationManager: "
      << exc.what();
    return StatusCode::CONFIGURATION_ERROR;
  }
}

auto ConfigurationManager::mClear() noexcept -> int {
  try {
    json_configuration = value(nullptr);
    ConnectionManager::Reset();
    return StatusCode::SUCCESS;
  } catch (const caosdb::exceptions::GenericException &exc) {
    return exc.GetCode();
  } catch (const std::exception &exc) {
    CAOSDB_LOG_ERROR(logger_name)
      << "Unknown error during the reset of the ConfigurationManager: "
      << exc.what();
    return StatusCode::CONFIGURATION_ERROR;
  }
}

auto ConfigurationManager::mLoadSingleJSONConfiguration(const path &json_file)
  -> void {
  if (!json_configuration.is_null()) {
    throw ConfigurationError("This CaosDB client has already been configured.");
  }
  if (!exists(json_file)) {
    throw ConfigurationError("Configuration file does not exist.");
  }

  json_configuration = load_json_file(json_file);
}

auto ConfigurationManager::mGetConnectionConfiguration(
  const std::string &name) const -> std::unique_ptr<ConnectionConfiguration> {
  auto connection_json = GetConnection(name);
  return connection_configuration_helper.CreateConnectionConfiguration(
    connection_json);
}

auto ConfigurationManager::mGetDefaultConnectionName() const -> std::string {
  auto connections = GetConnections();
  if (connections.contains("default")) {
    auto default_connection = connections.at("default");
    if (default_connection.is_object()) {
      // the name is actually "default"
      return std::string("default");
    } else {
      assert(default_connection.is_string());
      auto default_connection_name = default_connection.as_string();
      // return the string value of connections.default
      return std::string(default_connection_name.c_str());
    }
  }
  if (connections.size() == 1) {
    // return the key of the first and only sub-element of connections.
    return connections.begin()->key().to_string();
  }
  throw ConfigurationError("Could not determine the default connection.");
}

auto ConfigurationManager::GetConfiguration() const -> const object & {
  if (json_configuration.is_null()) {
    throw ConfigurationError("This CaosDB client has not been configured.");
  }
  assert(json_configuration.is_object());
  return json_configuration.as_object();
}

auto ConfigurationManager::GetConnections() const -> const object & {
  const auto &configuration = GetConfiguration();
  if (!configuration.contains("connections")) {
    throw ConfigurationError(
      "This CaosDB client hasn't any configured connections.");
  }
  const auto &connections_value = configuration.at("connections");
  if (connections_value.is_null()) {
    throw ConfigurationError(
      "This CaosDB client hasn't any configured connections.");
  }
  assert(connections_value.is_object());
  const auto &connections_object = connections_value.as_object();
  if (connections_object.empty()) {
    throw ConfigurationError(
      "This CaosDB client hasn't any configured connections.");
  }
  return connections_object;
}

auto ConfigurationManager::GetConnection(const std::string &name) const
  -> const object & {
  const auto &connections = GetConnections();
  if (connections.contains(name)) {
    const auto &result_connection = connections.at(name);
    assert(result_connection.is_object());
    return result_connection.as_object();
  }
  throw ConfigurationError("The connection '" + name +
                           "' has not been defined.");
}

// TODO(tf) This has apparently a cognitive complexity of 34>25 (threshold).
auto ConfigurationManager::InitializeDefaults() -> int { // NOLINT

  // find the configuration file...
  std::unique_ptr<path> configuration_file_path;
  for (const std::string &configuration_file :
       caosdb::LIBCAOSDB_CONFIGURATION_FILES_PRECEDENCE) {
    if (configuration_file == "$CAOSDB_CLIENT_CONFIGURATION") {
      // user specified a file via the environment variable
      // TODO(tf) make this thread-secure (concurrency-mt-unsafe)
      const auto *from_env_var =
        getenv("CAOSDB_CLIENT_CONFIGURATION"); // NOLINT
      if (from_env_var != nullptr) {
        configuration_file_path = std::make_unique<path>(from_env_var);
        if (exists(*configuration_file_path)) {
          break;
        } else {
          configuration_file_path = nullptr;
          // TODO(tf) log warning: "Configuration file under
          // $CAOSDB_CLIENT_CONFIGURATION does not exist.
        }
      }
    } else {
      // check standard locations
      configuration_file_path = std::make_unique<path>();
      const path raw(configuration_file);
      // resolve home directory
      for (auto segment = raw.begin(); segment != raw.end(); ++segment) {
        if (segment->string() == "$HOME") {
          path expanded_home(get_home_directory());
          *configuration_file_path /= expanded_home;
        } else {
          *configuration_file_path /= *segment;
        }
      }
      if (exists(*configuration_file_path)) {
        break;
      } else {
        configuration_file_path = nullptr;
      }
    }
  }

  // ... and use the configuration file
  if (configuration_file_path != nullptr) {
    mLoadSingleJSONConfiguration(*configuration_file_path);
  }

  if (this->json_configuration.is_object() &&
      this->json_configuration.as_object().contains("logging")) {
    LoggingConfiguration logging_configuration =
      logging_configuration_helper.CreateLoggingConfiguration(
        json_configuration.at("logging").as_object());
    logging::initialize_logging(logging_configuration);
  } else {
    logging::initialize_logging_defaults();
    CAOSDB_LOG_INFO(logger_name) << "No logging configuration has been found. "
                                    "We are using the default configuration";
  }

  if (configuration_file_path != nullptr &&
      this->json_configuration.is_object()) {
    CAOSDB_LOG_INFO(logger_name) << "Loaded configuration from "
                                 << *(configuration_file_path.get()) << ".";
  }

  return 0;
}

const int IS_INITIALIZED = ConfigurationManager::Reset();

} // namespace caosdb::configuration
